# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from ansible.module_utils.basic import AnsibleModule

import yaml


DOCUMENTATION = '''
---
module: metalsmith_deployment_defaults
short_description: Transform instances list data for metalsmith_instances
author: "Steve Baker (@stevebaker)"
description:
  - Takes a list of instances from the metalsmith_deployment role
    and a dict of defaults and transforms that to the format required
    by the metalsmith_instances module.
options:
  instances:
    description:
      - List of node description dicts to perform operations on (in the
        metalsmith_deployment instances format)
    type: list
    default: []
    elements: dict
  defaults:
    description:
      - Dict of defaults to use for missing values. Keys correspond to the
        metalsmith_deployment instances format.
    type: dict
'''


def transform(module, instances, defaults):
    mi = []

    def value(src, key, dest, to_key=None):
        if not to_key:
            to_key = key
        value = src.get(key, defaults.get(key))
        if value:
            dest[to_key] = value

    for src in instances:
        dest = {'image': {}}
        value(src, 'hostname', dest)
        value(src, 'candidates', dest)
        value(src, 'nics', dest)
        value(src, 'netboot', dest)
        value(src, 'root_size', dest, 'root_size_gb')
        value(src, 'swap_size', dest, 'swap_size_mb')
        value(src, 'capabilities', dest)
        value(src, 'traits', dest)
        value(src, 'resource_class', dest)
        value(src, 'conductor_group', dest)
        value(src, 'user_name', dest)
        image = dest['image']
        value(src, 'image', image, 'href')
        value(src, 'image_checksum', image, 'checksum')
        value(src, 'image_kernel', image, 'kernel')
        value(src, 'image_ramdisk', image, 'ramdisk')
        value(src, 'config_drive', dest)

        # keys in metalsmith_instances not currently in metalsmith_deployment:
        # passwordless_sudo

        # keys in metalsmith_deployment not currently in metalsmith_instances:
        # extra_args (CLI args cannot translate to the python lib,
        #             but they are mainly for auth and output formatting apart
        #             from --dry-run)
        if 'extra_args' in src:
            module.fail_json(
                changed=False,
                msg="extra_args is no longer supported"
            )

        # state (metalsmith_instances has a single state attribute for every
        #        instance)
        if 'state' in src:
            module.fail_json(
                changed=False,
                msg="Per-instance state is no longer supported, "
                    "use variable metalsmith_state"
            )

        # source keys could be a string or a list of strings
        # and the strings could be a path to a public key or the key contents.
        # Normalize this to a list of key contents
        keys = []
        source_keys = src.get('ssh_public_keys')
        if source_keys:
            if isinstance(source_keys, str):
                source_keys = [source_keys]
            for source_key in source_keys:
                if os.path.isfile(source_key):
                    with open(source_key) as f:
                        source_key = f.read()
                keys.append(source_key)
        if keys:
            dest['ssh_public_keys'] = keys

        mi.append(dest)

    module.exit_json(
        changed=False,
        msg="{} instances transformed".format(len(mi)),
        instances=mi
    )
    return mi


def main():
    module = AnsibleModule(
        argument_spec=yaml.safe_load(DOCUMENTATION)['options'],
        supports_check_mode=False,
    )

    instances = module.params['instances']
    defaults = module.params['defaults']
    transform(module, instances, defaults)


if __name__ == '__main__':
    main()



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from openstack import connection
from openstack import exceptions as os_exc

from metalsmith import _instance
from metalsmith import _network_metadata
from metalsmith import _nics
from metalsmith import _scheduler
from metalsmith import _utils
from metalsmith import exceptions
from metalsmith import instance_config
from metalsmith import sources


LOG = logging.getLogger(__name__)

_CREATED_PORTS = 'metalsmith_created_ports'
_ATTACHED_PORTS = 'metalsmith_attached_ports'
_PRESERVE_INSTANCE_INFO_KEYS = {'capabilities', 'traits'}


class Provisioner(object):
    """API to deploy/undeploy nodes with OpenStack.

    :param session: `Session` object (from ``keystoneauth``) to use when
        making API requests. Mutually exclusive with **cloud_region**.
    :param cloud_region: cloud configuration object (from ``openstacksdk``)
        to use when making API requests. Mutually exclusive with **session**.
    :param dry_run: boolean value, set to ``True`` to prevent any API calls
        from being actually made.

    :ivar connection: `openstacksdk` `Connection` object used for accessing
        OpenStack API during provisioning.
    """

    def __init__(self, session=None, cloud_region=None, dry_run=False):
        if cloud_region is None:
            if session is None:
                raise TypeError('Either session or cloud_region must '
                                'be provided')
            self.connection = connection.Connection(session=session)
        elif session is not None:
            raise TypeError('Either session or cloud_region must be provided, '
                            'but not both')
        else:
            self.connection = connection.Connection(config=cloud_region)

        self._dry_run = dry_run

    def reserve_node(self, resource_class, conductor_group=None,
                     capabilities=None, traits=None, candidates=None,
                     predicate=None, hostname=None):
        """Find and reserve a suitable node.

        Example::

         node = provisioner.reserve_node("compute",
                                         capabilities={"boot_mode": "uefi"})

        :param resource_class: Requested resource class.
        :param conductor_group: Conductor group to pick the nodes from.
            Value ``None`` means any group, use empty string "" for nodes
            from the default group.
        :param capabilities: Requested capabilities as a dict.
        :param traits: Requested traits as a list of strings.
        :param candidates: List of nodes (UUIDs, names or `Node` objects)
            to pick from. The filters (for resource class and capabilities)
            are still applied to the provided list. The order in which
            the nodes are considered is retained.
        :param predicate: Custom predicate to run on nodes. A callable that
            accepts a node and returns ``True`` if it should be included,
            ``False`` otherwise. Any exceptions are propagated to the caller.
        :param hostname: Hostname to assign to the instance. Defaults to the
            node's name or UUID.
        :return: reserved `Node` object.
        :raises: :py:class:`metalsmith.exceptions.ReservationFailed`
        """
        capabilities = capabilities or {}
        _utils.check_hostname(hostname)

        if candidates or capabilities or conductor_group or predicate:
            # Predicates, capabilities and conductor groups are not supported
            # by the allocation API natively, so we need to use prefiltering.
            candidates = self._prefilter_nodes(resource_class,
                                               conductor_group=conductor_group,
                                               capabilities=capabilities,
                                               candidates=candidates,
                                               predicate=predicate)

        node = self._reserve_node(resource_class, hostname=hostname,
                                  candidates=candidates, traits=traits,
                                  capabilities=capabilities)[0]
        return node

    def _prefilter_nodes(self, resource_class, conductor_group, capabilities,
                         candidates, predicate):
        """Build a list of candidate nodes for allocation."""
        if candidates:
            try:
                nodes = [self._get_node(node) for node in candidates]
            except os_exc.ResourceNotFound as exc:
                raise exceptions.InvalidNode(str(exc))
        else:
            nodes = list(self.connection.baremetal.nodes(
                details=True,
                associated=False,
                provision_state='available',
                maintenance=False,
                resource_class=resource_class,
                conductor_group=conductor_group))
            if not nodes:
                raise exceptions.NodesNotFound(resource_class, conductor_group)

        filters = [
            _scheduler.NodeTypeFilter(resource_class, conductor_group),
        ]
        if capabilities:
            filters.append(_scheduler.CapabilitiesFilter(capabilities))
        if predicate is not None:
            filters.append(_scheduler.CustomPredicateFilter(predicate))

        return _scheduler.run_filters(filters, nodes)

    def _reserve_node(self, resource_class, hostname=None, candidates=None,
                      traits=None, capabilities=None,
                      update_instance_info=True):
        """Create an allocation with given parameters."""
        if candidates:
            candidates = [
                (node.id if not isinstance(node, str) else node)
                for node in candidates
            ]

        LOG.debug('Creating an allocation for resource class %(rsc)s '
                  'with traits %(traits)s and candidate nodes %(candidates)s',
                  {'rsc': resource_class, 'traits': traits,
                   'candidates': candidates})
        try:
            allocation = self.connection.baremetal.create_allocation(
                name=hostname, candidate_nodes=candidates,
                resource_class=resource_class, traits=traits)
        except os_exc.SDKException as exc:
            # Re-raise the expected exception class
            raise exceptions.ReservationFailed(
                'Failed to create an allocation: %s' % exc)

        node = None
        try:
            try:
                allocation = self.connection.baremetal.wait_for_allocation(
                    allocation)
            except os_exc.SDKException as exc:
                # Re-raise the expected exception class
                raise exceptions.ReservationFailed(
                    'Failed to reserve a node: %s' % exc)

            LOG.info('Successful allocation %(alloc)s for host %(host)s',
                     {'alloc': allocation, 'host': hostname})
            node = self.connection.baremetal.get_node(allocation.node_id)

            if update_instance_info:
                node = self._patch_reserved_node(node, allocation, hostname,
                                                 capabilities)
        except Exception as exc:
            with _utils.reraise_os_exc(
                    exceptions.ReservationFailed,
                    'Failed to delete failed allocation') as expected:
                LOG.error('Processing allocation %(alloc)s for node %(node)s '
                          'failed: %(exc)s; deleting allocation',
                          {'alloc': _utils.log_res(allocation),
                           'node': _utils.log_res(node), 'exc': exc},
                          exc_info=not expected)
                self.connection.baremetal.delete_allocation(allocation)

        LOG.debug('Reserved node: %s', node)
        return node, allocation

    def _patch_reserved_node(self, node, allocation, hostname, capabilities):
        """Make required updates on a newly reserved node."""
        if capabilities:
            patch = [{'path': '/instance_info/capabilities',
                      'op': 'add', 'value': capabilities}]
            LOG.debug('Patching reserved node %(node)s with %(patch)s',
                      {'node': _utils.log_res(node), 'patch': patch})
            return self.connection.baremetal.patch_node(node, patch)
        else:
            return node

    def _check_node_for_deploy(self, node, hostname):
        """Check that node is ready and reserve it if needed.

        These checks are done outside of the try..except block in
        ``provision_node``, so that we don't touch nodes that fail it at all.
        Particularly, we don't want to try clean up nodes that were not
        reserved by us or are in maintenance mode.
        """
        if node.is_maintenance:
            raise exceptions.InvalidNode('Refusing to deploy on node %(node)s '
                                         'which is in maintenance mode due to '
                                         '%(reason)s' %
                                         {'node': _utils.log_res(node),
                                          'reason': node.maintenance_reason})

        allocation = None

        # Make sure the hostname does not correspond to an existing allocation
        # for another node.
        if hostname is not None:
            allocation = self._check_allocation_for_hostname(node, hostname)

        if node.allocation_id:
            if allocation is None:
                # Previously created allocation, verify/update it
                allocation = self._check_and_update_allocation_for_node(
                    node, hostname)
        elif node.instance_id:
            # Old-style reservations with instance_uuid==node.uuid
            if node.instance_id != node.id:
                raise exceptions.InvalidNode(
                    'Node %(node)s already reserved by instance %(inst)s '
                    'outside of metalsmith, cannot deploy on it' %
                    {'node': _utils.log_res(node), 'inst': node.instance_id})
            elif hostname:
                # We have no way to update hostname without allocations
                raise exceptions.InvalidNode(
                    'Node %s does not use allocations, cannot update '
                    'hostname for it' % _utils.log_res(node))
        else:
            # Node is not reserved at all - reserve it
            if not node.resource_class:
                raise exceptions.InvalidNode(
                    'Cannot create an allocation for node %s that '
                    'does not have a resource class set'
                    % _utils.log_res(node))

            if not self._dry_run:
                if not hostname:
                    hostname = _utils.default_hostname(node)
                LOG.debug('Node %(node)s is not reserved yet, reserving for '
                          'hostname %(host)s',
                          {'node': _utils.log_res(node),
                           'host': hostname})
                # Not updating instance_info since it will be updated later
                node, allocation = self._reserve_node(
                    node.resource_class,
                    hostname=hostname,
                    candidates=[node.id],
                    update_instance_info=False)

        return node, allocation

    def _check_allocation_for_hostname(self, node, hostname):
        try:
            allocation = self.connection.baremetal.get_allocation(
                hostname)
        except os_exc.ResourceNotFound:
            return

        if allocation.node_id and allocation.node_id != node.id:
            raise ValueError("The following node already uses "
                             "hostname %(host)s: %(node)s" %
                             {'host': hostname,
                              'node': allocation.node_id})
        else:
            return allocation

    def _check_and_update_allocation_for_node(self, node, hostname=None):
        # No allocation with given hostname, find one corresponding to the
        # node.
        allocation = self.connection.baremetal.get_allocation(
            node.allocation_id)
        if allocation.name and hostname and allocation.name != hostname:
            # Prevent updating of an existing hostname, since we don't
            # understand the intention
            raise exceptions.InvalidNode(
                "Allocation %(alloc)s associated with node %(node)s "
                "uses hostname %(old)s that does not match the expected "
                "hostname %(new)s" %
                {'alloc': _utils.log_res(allocation),
                 'node': _utils.log_res(node),
                 'old': allocation.name,
                 'new': hostname})
        elif not allocation.name and not self._dry_run:
            if not hostname:
                hostname = _utils.default_hostname(node)
            # Set the hostname that was not set in reserve_node.
            LOG.debug('Updating allocation %(alloc)s for node '
                      '%(node)s with hostname %(host)s',
                      {'alloc': _utils.log_res(allocation),
                       'node': _utils.log_res(node),
                       'host': hostname})
            allocation = self.connection.baremetal.update_allocation(
                allocation, name=hostname)

        return allocation

    def provision_node(self, node, image, nics=None, root_size_gb=None,
                       swap_size_mb=None, config=None, hostname=None,
                       netboot=False, capabilities=None, traits=None,
                       wait=None, clean_up_on_failure=True):
        """Provision the node with the given image.

        Example::

         provisioner.provision_node("compute-1", "centos",
                                    nics=[{"network": "private"},
                                          {"network": "external"}],
                                    root_size_gb=50,
                                    wait=3600)

        :param node: Node object, UUID or name. Will be reserved first, if
            not reserved already. Must be in the "available" state with
            maintenance mode off.
        :param image: Image source - one of :mod:`~metalsmith.sources`,
            `Image` name or UUID.
        :param nics: List of virtual NICs to attach to physical ports.
            Each item is a dict with a key describing the type of the NIC:

            * ``{"port": "<port name or ID>"}`` to use the provided pre-created
              port.
            * ``{"network": "<network name or ID>"}`` to create a port on the
              provided network. Optionally, a ``fixed_ip`` argument can be used
              to specify an IP address.
            * ``{"subnet": "<subnet name or ID>"}`` to create a port with an IP
              address from the provided subnet. The network is determined from
              the subnet.

        :param root_size_gb: The size of the root partition. By default
            the value of the local_gb property is used.
        :param swap_size_mb: The size of the swap partition. It's an error
            to specify it for a whole disk image.
        :param config: configuration to pass to the instance, one of
            objects from :py:mod:`metalsmith.instance_config`.
        :param hostname: Hostname to assign to the instance. If provided,
            overrides the ``hostname`` passed to ``reserve_node``.
        :param netboot: Whether to use networking boot for final instances.
        :param capabilities: Requested capabilities of the node. If present,
            overwrites the capabilities set by :meth:`reserve_node`.
            Note that the capabilities are not checked against the ones
            provided by the node - use :meth:`reserve_node` for that.
        :param traits: Requested traits of the node. If present, overwrites
            the traits set by :meth:`reserve_node`. Note that the traits are
            not checked against the ones provided by the node - use
            :meth:`reserve_node` for that.
        :param wait: How many seconds to wait for the deployment to finish,
            None to return immediately.
        :param clean_up_on_failure: If True, then on failure the node is
            cleared of instance information, VIFs are detached, created ports
            and allocations are deleted.
        :return: :py:class:`metalsmith.Instance` object with the current
            status of provisioning. If ``wait`` is not ``None``, provisioning
            is already finished.
        :raises: :py:class:`metalsmith.exceptions.Error`
        """
        if config is None:
            config = instance_config.GenericConfig()
        if isinstance(image, str):
            image = sources.GlanceImage(image)

        _utils.check_hostname(hostname)

        try:
            node = self._get_node(node)
        except Exception as exc:
            raise exceptions.InvalidNode('Cannot find node %(node)s: %(exc)s' %
                                         {'node': node, 'exc': exc})

        node, allocation = self._check_node_for_deploy(node, hostname)
        nics = _nics.NICs(self.connection, node, nics,
                          hostname=allocation and allocation.name or None)

        try:
            root_size_gb = _utils.get_root_disk(root_size_gb, node)

            image._validate(self.connection, root_size_gb)

            nics.validate()

            if capabilities is None:
                capabilities = node.instance_info.get('capabilities') or {}

            if self._dry_run:
                LOG.warning('Dry run, not provisioning node %s',
                            _utils.log_res(node))
                return node

            nics.create_and_attach_ports()

            capabilities['boot_option'] = 'netboot' if netboot else 'local'

            instance_info = self._clean_instance_info(node.instance_info)
            if root_size_gb is not None:
                instance_info['root_gb'] = root_size_gb
            instance_info['capabilities'] = capabilities
            if hostname:
                instance_info['display_name'] = hostname

            extra = node.extra.copy()
            extra[_CREATED_PORTS] = nics.created_ports
            extra[_ATTACHED_PORTS] = nics.attached_ports
            instance_info.update(image._node_updates(self.connection))
            if traits is not None:
                instance_info['traits'] = traits
            if swap_size_mb is not None:
                instance_info['swap_mb'] = swap_size_mb

            LOG.debug('Updating node %(node)s with instance info %(iinfo)s '
                      'and extras %(extra)s', {'node': _utils.log_res(node),
                                               'iinfo': instance_info,
                                               'extra': extra})
            node = self.connection.baremetal.update_node(
                node, instance_info=instance_info, extra=extra)
            self.connection.baremetal.validate_node(node)

            network_data = _network_metadata.create_network_metadata(
                self.connection, node.extra.get(_ATTACHED_PORTS))

            LOG.debug('Generating a configdrive for node %s',
                      _utils.log_res(node))
            cd = config.generate(node, _utils.hostname_for(node, allocation),
                                 network_data)
            LOG.debug('Starting provisioning of node %s', _utils.log_res(node))
            self.connection.baremetal.set_node_provision_state(
                node, 'active', config_drive=cd)
        except Exception:
            with _utils.reraise_os_exc(
                    exceptions.DeploymentFailed) as expected:
                if clean_up_on_failure:
                    LOG.error('Deploy attempt failed on node %s, cleaning up',
                              _utils.log_res(node), exc_info=not expected)
                    self._clean_up(node, nics=nics)

        LOG.info('Provisioning started on node %s', _utils.log_res(node))

        if wait is not None:
            LOG.debug('Waiting for node %(node)s to reach state active '
                      'with timeout %(timeout)s',
                      {'node': _utils.log_res(node), 'timeout': wait})
            instance = self.wait_for_provisioning([node], timeout=wait)[0]
            LOG.info('Deploy succeeded on node %s', _utils.log_res(node))
        else:
            # Update the node to return it's latest state
            node = self.connection.baremetal.get_node(node.id)
            instance = _instance.Instance(self.connection, node, allocation)

        return instance

    def wait_for_provisioning(self, nodes, timeout=None):
        """Wait for nodes to be provisioned.

        Loops until all nodes finish provisioning.

        :param nodes: List of nodes (UUID, name, `Node` object or
            :py:class:`metalsmith.Instance`).
        :param timeout: How much time (in seconds) to wait for all nodes
            to finish provisioning. If ``None`` (the default), wait forever
            (more precisely, until the operation times out on server side).
        :return: List of updated :py:class:`metalsmith.Instance` objects if
            all succeeded.
        :raises: :py:class:`metalsmith.exceptions.DeploymentFailed`
            if deployment fails or times out.
        :raises: :py:class:`metalsmith.exceptions.InstanceNotFound`
            if requested nodes cannot be found.
        """
        nodes = [self._find_node_and_allocation(n)[0] for n in nodes]
        try:
            nodes = self.connection.baremetal.wait_for_nodes_provision_state(
                nodes, 'active', timeout=timeout)
        except os_exc.ResourceTimeout as exc:
            raise exceptions.DeploymentTimeout(str(exc))
        except os_exc.SDKException as exc:
            raise exceptions.DeploymentFailed(str(exc))

        # Using _get_instance in case the deployment started by something
        # external that uses allocations.
        return [self._get_instance(node) for node in nodes]

    def _clean_instance_info(self, instance_info):
        return {key: value
                for key, value in instance_info.items()
                if key in _PRESERVE_INSTANCE_INFO_KEYS}

    def _clean_up(self, node, nics=None, remove_instance_info=True):
        if nics is None:
            created_ports = node.extra.get(_CREATED_PORTS, [])
            attached_ports = node.extra.get(_ATTACHED_PORTS, [])
            _nics.detach_and_delete_ports(self.connection, node,
                                          created_ports, attached_ports)
        else:
            nics.detach_and_delete_ports()

        extra = node.extra.copy()
        for item in (_CREATED_PORTS, _ATTACHED_PORTS):
            extra.pop(item, None)

        kwargs = {}
        if node.allocation_id and node.provision_state != 'active':
            # Try to remove allocation (it will fail for active nodes)
            LOG.debug('Trying to remove allocation %(alloc)s for node '
                      '%(node)s', {'alloc': node.allocation_id,
                                   'node': _utils.log_res(node)})
            try:
                self.connection.baremetal.delete_allocation(node.allocation_id)
            except Exception as exc:
                LOG.debug('Failed to remove allocation %(alloc)s for %(node)s:'
                          ' %(exc)s',
                          {'alloc': node.allocation_id,
                           'node': _utils.log_res(node), 'exc': exc})
        elif not node.allocation_id:
            # Old-style reservations have to be cleared explicitly
            kwargs['instance_id'] = None

        try:
            if remove_instance_info:
                LOG.debug('Updating node %(node)s with empty instance info '
                          '(was %(iinfo)s) and extras %(extra)s',
                          {'node': _utils.log_res(node),
                           'iinfo': node.instance_info,
                           'extra': extra})
                self.connection.baremetal.update_node(
                    node, instance_info={}, extra=extra, **kwargs)
            else:
                LOG.debug('Updating node %(node)s with extras %(extra)s',
                          {'node': _utils.log_res(node), 'extra': extra})
                self.connection.baremetal.update_node(
                    node, extra=extra, **kwargs)
        except Exception as exc:
            LOG.debug('Failed to clear node %(node)s extra: %(exc)s',
                      {'node': _utils.log_res(node), 'exc': exc})

    def unprovision_node(self, node, wait=None):
        """Unprovision a previously provisioned node.

        :param node: `Node` object, :py:class:`metalsmith.Instance`,
            hostname, UUID or node name.
        :param wait: How many seconds to wait for the process to finish,
            None to return immediately.
        :return: the latest `Node` object.
        :raises: :py:class:`metalsmith.exceptions.DeploymentFailed`
            if undeployment fails.
        :raises: :py:class:`metalsmith.exceptions.DeploymentTimeout`
            if undeployment times out.
        :raises: :py:class:`metalsmith.exceptions.InstanceNotFound`
            if requested node cannot be found.
        """
        node = self._find_node_and_allocation(node)[0]
        if self._dry_run:
            LOG.warning("Dry run, not unprovisioning")
            return

        self._clean_up(node, remove_instance_info=False)
        try:
            node = self.connection.baremetal.set_node_provision_state(
                node, 'deleted', wait=False)

            LOG.info('Deleting started for node %s', _utils.log_res(node))

            if wait is None:
                return node

            node = self.connection.baremetal.wait_for_nodes_provision_state(
                [node], 'available', timeout=wait)[0]
        except os_exc.ResourceTimeout as exc:
            raise exceptions.DeploymentTimeout(str(exc))
        except os_exc.SDKException as exc:
            raise exceptions.DeploymentFailed(str(exc))

        LOG.info('Node %s undeployed successfully', _utils.log_res(node))
        return node

    def show_instance(self, instance_id):
        """Show information about instance.

        :param instance_id: hostname, UUID or node name.
        :return: :py:class:`metalsmith.Instance` object.
        :raises: :py:class:`metalsmith.exceptions.InstanceNotFound`
            if the instance is not a valid instance.
        """
        return self.show_instances([instance_id])[0]

    def show_instances(self, instances):
        """Show information about instance.

        More efficient than calling :meth:`show_instance` in a loop, because
        it caches the node list.

        :param instances: list of hostnames, UUIDs or node names.
        :return: list of :py:class:`metalsmith.Instance` objects in the same
            order as ``instances``.
        :raises: :py:class:`metalsmith.exceptions.InstanceNotFound`
            if one of the instances cannot be found or the found node is
            not a valid instance.
        """
        result = [self._get_instance(inst) for inst in instances]
        # NOTE(dtantsur): do not accept node names as valid instances if they
        # are not deployed or being deployed.
        missing = [inst for (res, inst) in zip(result, instances)
                   if res.state == _instance.InstanceState.UNKNOWN]
        if missing:
            raise exceptions.InstanceNotFound(
                "Node(s)/instance(s) %s are not valid instances"
                % ', '.join(map(str, missing)))
        return result

    def list_instances(self):
        """List instances deployed by metalsmith.

        :return: list of :py:class:`metalsmith.Instance` objects.
        """
        nodes = self.connection.baremetal.nodes(associated=True, details=True)
        instances = [i for i in map(self._get_instance, nodes)
                     if i.state != _instance.InstanceState.UNKNOWN]
        return instances

    def _get_node(self, node, refresh=False):
        """A helper to find and return a node."""
        if isinstance(node, str):
            return self.connection.baremetal.get_node(node)
        elif hasattr(node, 'node'):
            # Instance object
            node = node.node
        else:
            node = node

        if refresh:
            return self.connection.baremetal.get_node(node.id)
        else:
            return node

    def _find_node_and_allocation(self, node, refresh=False):
        try:
            if (not isinstance(node, str)
                    or not _utils.is_hostname_safe(node)):
                return self._get_node(node, refresh=refresh), None

            try:
                allocation = self.connection.baremetal.get_allocation(node)
            except os_exc.ResourceNotFound:
                return self._get_node(node, refresh=refresh), None
        except os_exc.ResourceNotFound as exc:
            raise exceptions.InstanceNotFound(str(exc))

        if allocation.node_id:
            try:
                return (self.connection.baremetal.get_node(allocation.node_id),
                        allocation)
            except os_exc.ResourceNotFound:
                raise exceptions.InstanceNotFound(
                    'Node %(node)s associated with allocation '
                    '%(alloc)s was not found' %
                    {'node': allocation.node_id,
                     'alloc': allocation.id})
        else:
            raise exceptions.InstanceNotFound(
                'Allocation %s exists but is not associated '
                'with a node' % node)

    def _get_instance(self, ident):
        node, allocation = self._find_node_and_allocation(ident)
        if allocation is None and node.allocation_id:
            try:
                allocation = self.connection.baremetal.get_allocation(
                    node.allocation_id)
            except os_exc.ResourceNotFound as exc:
                raise exceptions.InstanceNotFound(str(exc))

        return _instance.Instance(self.connection, node,
                                  allocation=allocation)



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ipaddress

from openstack import exceptions as sdk_exc

from metalsmith import exceptions


def create_network_metadata(connection, attached_ports):
    network_data = {}
    if not attached_ports:
        return network_data

    links = network_data.setdefault('links', [])
    networks = network_data.setdefault('networks', [])
    services = network_data.setdefault('services', [])

    for attached_port in attached_ports:
        try:
            port = connection.network.get_port(attached_port)
            net = connection.network.get_network(port.network_id)
            subnets = [connection.network.get_subnet(x['subnet_id'])
                       for x in port.fixed_ips]
            subnets_by_id = {x.id: x for x in subnets}
        except sdk_exc.SDKException as exc:
            raise exceptions.NetworkResourceNotFound(
                'Cannot find network resource: %s' % exc)

        metadata_add_links(links, port, net)
        metadata_add_services(services, subnets)
        for idx, fixed_ip in enumerate(port.fixed_ips):
            subnet = subnets_by_id[fixed_ip['subnet_id']]
            metadata_add_network(networks, idx, fixed_ip, port, net, subnet)

    return network_data


def metadata_add_links(links, port, network):
    links.append({'id': port.id,
                  'type': 'phy',
                  'mtu': network.mtu,
                  'ethernet_mac_address': port.mac_address})


def metadata_add_services(services, subnets):
    for subnet in subnets:
        for dns_nameserver in subnet.dns_nameservers:
            services.append({'type': 'dns',
                             'address': dns_nameserver})


def metadata_add_network(networks, idx, fixed_ip, port, network, subnet):
    ip_net = ipaddress.ip_network(subnet.cidr)

    net_data = {'id': network.name + str(idx),
                'network_id': network.id,
                'link': port.id,
                'ip_address': fixed_ip['ip_address'],
                'netmask': str(ip_net.netmask)}

    if subnet.ip_version == 4:
        net_data['type'] = 'ipv4_dhcp' if subnet.is_dhcp_enabled else 'ipv4'
    elif subnet.ip_version == 6:
        net_data['type'] = ('ipv6_{}'.format(subnet.ipv6_address_mode)
                            if subnet.ipv6_address_mode else 'ipv6')

    net_routes = net_data.setdefault('routes', [])
    for route in subnet.host_routes:
        ip_net = ipaddress.ip_network(route['destination'])
        net_routes.append({'network': str(ip_net.network_address),
                           'netmask': str(ip_net.netmask),
                           'gateway': route['nexthop']})

    # Services go in both "network" and toplevel.
    # Ref: https://docs.openstack.org/nova/latest/_downloads/9119ca7ac90aa2990e762c08baea3a36/network_data.json  # noqa
    net_services = net_data.setdefault('services', [])
    metadata_add_services(net_services, [subnet])

    networks.append(net_data)



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import json
import sys

import prettytable

from metalsmith import _utils


def _print(msg, **kwargs):
    print(msg % kwargs)


class NullFormat(object):
    """Formatting outputting nothing.

    Used implicitly with --quiet.
    """

    def __init__(self, columns=None, sort_column=None):
        self.columns = columns
        self.sort_column = sort_column

    def deploy(self, instance):
        pass

    def undeploy(self, node):
        pass


FIELDS = ['UUID', 'Node Name', 'Allocation UUID', 'Hostname',
          'State', 'IP Addresses']


class ValueFormat(NullFormat):
    """"Simple value formatter."""

    def deploy(self, instance):
        """Output result of the deploy."""
        self.show([instance])

    def undeploy(self, node):
        """Output result of undeploy."""
        if node.provision_state == 'available':
            message = "Successfully unprovisioned node %(node)s"
        else:
            message = "Unprovisioning started for node %(node)s"

        _print(message, node=_utils.log_res(node))

    def _iter_rows(self, instances):
        for instance in instances:
            if instance.is_deployed:
                ips = '\n'.join('%s=%s' % (net, ','.join(ips))
                                for net, ips in
                                instance.ip_addresses().items())
            else:
                ips = ''
            row = [instance.uuid, instance.node.name or '',
                   instance.allocation.id if instance.allocation else '',
                   instance.hostname or '', instance.state.name, ips]
            yield row

    def show(self, instances):
        allowed_columns = set(self.columns or FIELDS)
        rows = (collections.OrderedDict(zip(FIELDS, row))
                for row in self._iter_rows(instances))
        if self.sort_column:
            rows = sorted(rows, key=lambda row: row.get(self.sort_column))
        for row in rows:
            _print(' '.join(value if value is not None else ''
                            for key, value in row.items()
                            if key in allowed_columns))


class DefaultFormat(ValueFormat):
    """Human-readable formatter."""

    def show(self, instances):
        if not instances:
            _print('')  # Compatibility with openstackclient - one empty line
            return

        pt = prettytable.PrettyTable(field_names=FIELDS)
        pt.align = 'l'
        if self.sort_column:
            pt.sortby = self.sort_column

        for row in self._iter_rows(instances):
            pt.add_row(row)

        if self.columns:
            value = pt.get_string(fields=self.columns)
        else:
            value = pt.get_string()
        _print(value)


class JsonFormat(NullFormat):
    """JSON formatter."""

    def deploy(self, instance):
        """Output result of the deploy."""
        json.dump(instance.to_dict(), sys.stdout)

    def undeploy(self, node):
        """Output result of undeploy."""
        result = {
            'node': node.to_dict()
        }
        json.dump(result, sys.stdout)

    def show(self, instances):
        """Output instance statuses."""
        json.dump({instance.hostname: instance.to_dict()
                   for instance in instances}, sys.stdout)


FORMATS = {
    'default': DefaultFormat,
    'json': JsonFormat,
    'table': DefaultFormat,
    'value': ValueFormat,
}
"""Available formatters."""


DEFAULT_FORMAT = 'table'
"""Default formatter."""


NULL_FORMAT = NullFormat()
"""Formatter outputting nothing."""



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

from metalsmith import _scheduler
from metalsmith import exceptions


class TestRunFilters(unittest.TestCase):

    def setUp(self):
        super(TestRunFilters, self).setUp()
        self.nodes = [mock.Mock(spec=['id', 'name']) for _ in range(2)]

    def _filter(self, side_effect, fail=AssertionError('called fail')):
        fltr = mock.Mock(spec=_scheduler.Filter)
        fltr.side_effect = side_effect
        fltr.fail.side_effect = fail
        return fltr

    def test_no_filters(self):
        result = _scheduler.run_filters([], self.nodes)
        self.assertEqual(result, self.nodes)

    def test_all_filters_pass(self):
        filters = [self._filter([True, True]) for _ in range(3)]
        result = _scheduler.run_filters(filters, self.nodes)
        self.assertEqual(result, self.nodes)
        for fltr in filters:
            self.assertEqual([mock.call(n) for n in self.nodes],
                             fltr.call_args_list)
            self.assertFalse(fltr.fail.called)

    def test_one_node_filtered(self):
        filters = [self._filter([True, True]),
                   self._filter([False, True]),
                   self._filter([True])]
        result = _scheduler.run_filters(filters, self.nodes)
        self.assertEqual(result, self.nodes[1:2])
        for fltr in filters:
            self.assertFalse(fltr.fail.called)
        for fltr in filters[:2]:
            self.assertEqual([mock.call(n) for n in self.nodes],
                             fltr.call_args_list)
        filters[2].assert_called_once_with(self.nodes[1])

    def test_all_nodes_filtered(self):
        filters = [self._filter([True, True]),
                   self._filter([False, True]),
                   self._filter([False], fail=RuntimeError('failed'))]
        self.assertRaisesRegex(RuntimeError, 'failed',
                               _scheduler.run_filters,
                               filters, self.nodes)
        for fltr in filters[:2]:
            self.assertEqual([mock.call(n) for n in self.nodes],
                             fltr.call_args_list)
            self.assertFalse(fltr.fail.called)
        filters[2].assert_called_once_with(self.nodes[1])
        filters[2].fail.assert_called_once_with()


class TestCapabilitiesFilter(unittest.TestCase):

    def test_fail_no_capabilities(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute'})
        self.assertRaisesRegex(exceptions.CapabilitiesNotFound,
                               'No available nodes found with capabilities '
                               'profile=compute, existing capabilities: none',
                               fltr.fail)

    def test_nothing_requested_nothing_found(self):
        fltr = _scheduler.CapabilitiesFilter({})
        node = mock.Mock(properties={}, spec=['properties', 'name', 'id'])
        self.assertTrue(fltr(node))

    def test_matching_node(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute',
                                              'foo': 'bar'})
        node = mock.Mock(
            properties={'capabilities': 'foo:bar,profile:compute,answer:42'},
            spec=['properties', 'name', 'id'])
        self.assertTrue(fltr(node))

    def test_not_matching_node(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute',
                                              'foo': 'bar'})
        node = mock.Mock(
            properties={'capabilities': 'foo:bar,answer:42'},
            spec=['properties', 'name', 'id'])
        self.assertFalse(fltr(node))

    def test_fail_message(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute'})
        node = mock.Mock(
            properties={'capabilities': 'profile:control'},
            spec=['properties', 'name', 'id'])
        self.assertFalse(fltr(node))
        self.assertRaisesRegex(exceptions.CapabilitiesNotFound,
                               'No available nodes found with capabilities '
                               'profile=compute, existing capabilities: '
                               r'profile=control \(1 node\(s\)\)',
                               fltr.fail)

    def test_malformed_capabilities(self):
        fltr = _scheduler.CapabilitiesFilter({'profile': 'compute'})
        for cap in ['foo,profile:control', 42, 'a:b:c']:
            node = mock.Mock(properties={'capabilities': cap},
                             spec=['properties', 'name', 'id'])
            self.assertFalse(fltr(node))
        self.assertRaisesRegex(exceptions.CapabilitiesNotFound,
                               'No available nodes found with capabilities '
                               'profile=compute, existing capabilities: none',
                               fltr.fail)



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import logging

from metalsmith import _utils


LOG = logging.getLogger(__name__)


class GenericConfig(object):
    """Configuration of the target instance.

    The information attached to this object will be passed via a configdrive
    to the instance's first boot script (e.g. cloud-init).

    This class represents generic configuration compatible with most first-boot
    implementations. Use :py:class:`CloudInitConfig` for features specific to
    `cloud-init <https://cloudinit.readthedocs.io/>`_.

    :ivar ssh_keys: List of SSH public keys.
    :ivar user_data: User data as a string.
    :ivar meta_data: Dict of data to add to the generated ``meta_data``
    """

    def __init__(self, ssh_keys=None, user_data=None, meta_data=None):
        self.ssh_keys = ssh_keys or []
        self.user_data = user_data
        if meta_data and not isinstance(meta_data, dict):
            raise TypeError('Custom meta_data must be a dictionary, '
                            'got %r' % meta_data)
        self.meta_data = meta_data or {}

    def generate(self, node, hostname=None, network_data=None):
        """Generate the config drive information.

        :param node: `Node` object.
        :param hostname: Desired hostname (defaults to node's name or ID).
        :param network_data: Network metadata as dictionary
        :return: configdrive contents as a dictionary with keys:

            ``meta_data``
                meta data dictionary
            ``network_data``
                network data as dictionary
            ``user_data``
                user data as a string
        """
        if not hostname:
            hostname = _utils.default_hostname(node)

        # NOTE(dtantsur): CirrOS does not understand lists
        if isinstance(self.ssh_keys, list):
            ssh_keys = {str(i): v for i, v in enumerate(self.ssh_keys)}
        else:
            ssh_keys = self.ssh_keys

        meta_data = self.meta_data.copy()
        meta_data.update({
            'public_keys': ssh_keys,
            'uuid': node.id,
            'name': node.name,
            'hostname': hostname
        })
        meta_data.setdefault('launch_index', 0)
        meta_data.setdefault('availability_zone', '')
        meta_data.setdefault('files', [])
        meta_data.setdefault('meta', {})

        user_data = self.populate_user_data()

        data = {'meta_data': meta_data, 'user_data': user_data}

        if network_data:
            data['network_data'] = network_data

        return data

    def populate_user_data(self):
        """Get user data for this configuration.

        Can be overridden to provide additional features.

        :return: user data as a string.
        """
        return self.user_data


class CloudInitConfig(GenericConfig):
    """Configuration of the target instance using cloud-init.

    Compared to :class:`GenericConfig`, this adds support for managing users.

    :ivar ssh_keys: List of SSH public keys.
    :ivar user_data: Cloud-init cloud-config data as a dictionary.
    :ivar meta_data: Dict of data to add to the generated ``meta_data``
    """

    def __init__(self, ssh_keys=None, user_data=None, meta_data=None):
        if user_data is not None and not isinstance(user_data, dict):
            raise TypeError('Custom user data must be a dictionary for '
                            'CloudInitConfig, got %r' % user_data)
        super(CloudInitConfig, self).__init__(ssh_keys, user_data or {},
                                              meta_data=meta_data)
        self.users = []

    def add_user(self, name, admin=True, password_hash=None, sudo=False,
                 **kwargs):
        """Add a user to be created on first boot.

        :param name: user name.
        :param admin: whether to add the user to the admin group (wheel).
        :param password_hash: user password hash, if password authentication
            is expected.
        :param sudo: whether to allow the user sudo without password.
        :param kwargs: other arguments to pass.
        """
        kwargs['name'] = name
        if admin:
            kwargs.setdefault('groups', []).append('wheel')
        if password_hash:
            kwargs['passwd'] = password_hash
        if sudo:
            kwargs['sudo'] = 'ALL=(ALL) NOPASSWD:ALL'
        if self.ssh_keys:
            kwargs.setdefault('ssh_authorized_keys', self.ssh_keys)
        self.users.append(kwargs)

    def populate_user_data(self):
        """Get user data for this configuration.

        Takes the custom user data and appends requested users to it.

        :return: user data as a string.
        """
        if not isinstance(self.user_data, dict):
            raise TypeError('Custom user data must be a dictionary for '
                            'CloudInitConfig, got %r' % self.user_data)

        if self.users:
            user_data = copy.deepcopy(self.user_data)
            user_data.setdefault('users', []).extend(self.users)
        else:
            user_data = self.user_data

        if user_data:
            return "#cloud-config\n" + json.dumps(user_data)



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import imp
import os

from docutils import core
from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import Directive
from docutils.writers.html4css1 import Writer

import yaml


DOCYAML = yaml
DOCYAML.default_flow_style = False


class AnsibleAutoPluginDirective(Directive):
    directive_name = "ansibleautoplugin"
    has_content = True
    option_spec = {
        'module': rst.directives.unchanged,
        'role': rst.directives.unchanged,
        'documentation': rst.directives.unchanged,
        'examples': rst.directives.unchanged
    }

    @staticmethod
    def _render_html(source):
        return core.publish_parts(
            source=source,
            writer=Writer(),
            writer_name='html',
            settings_overrides={'no_system_messages': True}
        )

    def make_node(self, title, contents, content_type=None):
        section = self._section_block(title=title)
        if not content_type:
            # Doc section
            for content in contents['docs']:
                for paragraph in content.split('\n'):
                    retnode = nodes.paragraph()
                    retnode.append(self._raw_html_block(data=paragraph))
                    section.append(retnode)

            # Options Section
            options_list = nodes.field_list()
            options_section = self._section_block(title='Options')
            for key, value in contents['options'].items():
                options_list.append(
                    self._raw_fields(
                        data=value['description'],
                        field_name=key
                    )
                )
            else:
                options_section.append(options_list)
                section.append(options_section)

            # Authors Section
            authors_list = nodes.field_list()
            authors_list.append(
                self._raw_fields(
                    data=contents['author']
                )
            )
            authors_section = self._section_block(title='Authors')
            authors_section.append(authors_list)
            section.append(authors_section)

        elif content_type == 'yaml':
            for content in contents:
                section.append(
                    self._literal_block(
                        data=content,
                        dump_data=False
                    )
                )

        return section

    @staticmethod
    def load_module(filename):
        return imp.load_source('__ansible_module__', filename)

    @staticmethod
    def build_documentation(module):
        docs = DOCYAML.load(module.DOCUMENTATION)
        doc_data = dict()
        doc_data['docs'] = docs['description']
        doc_data['author'] = docs.get('author', list())
        doc_data['options'] = docs.get('options', dict())
        return doc_data

    @staticmethod
    def build_examples(module):
        examples = DOCYAML.load(module.EXAMPLES)
        return_examples = list()
        for example in examples:
            return_examples.append(DOCYAML.dump([example], sort_keys=False))
        return return_examples

    def _raw_html_block(self, data):
        html = self._render_html(source=data)
        return nodes.raw('', html['body'], format='html')

    def _raw_fields(self, data, field_name=''):
        body = nodes.field_body()
        if isinstance(data, list):
            for item in data:
                body.append(self._raw_html_block(data=item))
        else:
            body.append(self._raw_html_block(data=data))

        field = nodes.field()
        field.append(nodes.field_name(text=field_name))
        field.append(body)
        return field

    @staticmethod
    def _literal_block(data, language='yaml', dump_data=True):
        if dump_data:
            literal = nodes.literal_block(
                text=DOCYAML.dump(data)
            )
        else:
            literal = nodes.literal_block(text=data)
        literal['language'] = 'yaml'
        return literal

    @staticmethod
    def _section_block(title, text=None):
        section = nodes.section(
            title,
            nodes.title(text=title),
            ids=[nodes.make_id('-'.join(title))],
        )
        if text:
            section_body = nodes.field_body()
            section_body.append(nodes.paragraph(text=text))
            section.append(section_body)

        return section

    def _yaml_section(self, to_yaml_data, section_title, section_text=None):
        yaml_section = self._section_block(
            title=section_title,
            text=section_text
        )
        yaml_section.append(self._literal_block(data=to_yaml_data))
        return yaml_section

    def _run_role(self, role):
        section = self._section_block(
            title='Role Documentation',
            text='Welcome to the "{}" role documentation.'.format(
                os.path.basename(role)
            )
        )
        defaults_file = os.path.join(role, 'defaults', 'main.yml')
        if os.path.exists(defaults_file):
            with open(defaults_file) as f:
                role_defaults = DOCYAML.load(f.read())
            section.append(
                self._yaml_section(
                    to_yaml_data=role_defaults,
                    section_title='Role Defaults',
                    section_text='This section highlights all of the defaults'
                                 ' and variables set within the "{}"'
                                 ' role.'.format(os.path.basename(role))
                )
            )

        vars_path = os.path.join(role, 'vars')
        if os.path.exists(vars_path):
            for v_file in os.listdir(vars_path):
                vars_file = os.path.join(vars_path, v_file)
                with open(vars_file) as f:
                    vars_values = DOCYAML.load(f.read())
                section.append(
                    self._yaml_section(
                        to_yaml_data=vars_values,
                        section_title='Role Variables: {}'.format(v_file)
                    )
                )

        self.run_returns.append(section)

        # Document any libraries nested within the role
        library_path = os.path.join(role, 'library')
        if os.path.exists(library_path):
            self.options['documentation'] = True
            self.options['examples'] = True
            for lib in os.listdir(library_path):
                if lib.endswith('.py'):
                    self._run_module(
                        module=self.load_module(
                            filename=os.path.join(
                                library_path,
                                lib
                            )
                        ),
                        module_title='Embedded module: {}'.format(lib),
                        example_title='Examples for embedded module'
                    )

    def _run_module(self, module, module_title="Module Documentation",
                    example_title="Example Tasks"):
        if self.options.get('documentation'):
            docs = self.build_documentation(module=module)
            self.run_returns.append(
                self.make_node(
                    title=module_title,
                    contents=docs
                )
            )

        if self.options.get('examples'):
            examples = self.build_examples(module=module)
            self.run_returns.append(
                self.make_node(
                    title=example_title,
                    contents=examples,
                    content_type='yaml'
                )
            )

    def run(self):
        self.run_returns = list()

        if self.options.get('module'):
            module = self.load_module(filename=self.options['module'])
            self._run_module(module=module)

        if self.options.get('role'):
            self._run_role(role=self.options['role'])

        return self.run_returns


def setup(app):
    classes = [
        AnsibleAutoPluginDirective,
    ]
    for directive_class in classes:
        app.add_directive(directive_class.directive_name, directive_class)

    return {'version': '0.2'}



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
from unittest import mock

from metalsmith import instance_config


class TestGenericConfig(unittest.TestCase):
    CLASS = instance_config.GenericConfig

    def setUp(self):
        super(TestGenericConfig, self).setUp()
        self.node = mock.Mock(id='1234')
        self.node.name = 'node name'

    def _check(self, config, expected_metadata, expected_userdata=None,
               cloud_init=True, hostname=None, network_data=None,
               expected_network_data=None):
        expected_m = {'public_keys': {},
                      'uuid': self.node.id,
                      'name': self.node.name,
                      'hostname': self.node.id,
                      'launch_index': 0,
                      'availability_zone': '',
                      'files': [],
                      'meta': {}}
        expected_m.update(expected_metadata)

        result = config.generate(self.node, hostname, network_data)
        self.assertEqual(expected_m, result['meta_data'])

        user_data = result['user_data']
        if expected_userdata:
            self.assertIsNotNone(user_data)
            if cloud_init:
                header, user_data = user_data.split('\n', 1)
                self.assertEqual('#cloud-config', header)
            user_data = json.loads(user_data)
        self.assertEqual(expected_userdata, user_data)

        network_data = result.get('network_data')
        if expected_network_data:
            self.assertIsNotNone(network_data)
            self.assertEqual(expected_network_data, network_data)

    def test_default(self):
        config = self.CLASS()
        self._check(config, {})

    def test_name_as_hostname(self):
        self.node.name = 'example.com'
        config = self.CLASS()
        self._check(config, {'hostname': 'example.com'})

    def test_explicit_hostname(self):
        config = self.CLASS()
        self._check(config, {'hostname': 'example.com'},
                    hostname='example.com')

    def test_ssh_keys(self):
        config = self.CLASS(ssh_keys=['abc', 'def'])
        self._check(config, {'public_keys': {'0': 'abc', '1': 'def'}})

    def test_ssh_keys_as_dict(self):
        config = self.CLASS(ssh_keys={'default': 'abc'})
        self._check(config, {'public_keys': {'default': 'abc'}})

    def test_custom_user_data(self):
        config = self.CLASS(user_data='{"answer": 42}')
        self._check(config, {}, {"answer": 42}, cloud_init=False)

    def test_custom_metadata(self):
        config = self.CLASS(meta_data={"foo": "bar"})
        self._check(config, {"foo": "bar"}, cloud_init=False)

    def test_custom_metadata_not_dict(self):
        self.assertRaises(TypeError, self.CLASS, meta_data="foobar")

    def test_custom_network_data(self):
        config = self.CLASS()
        data = {'net': 'data'}
        self._check(config, {}, network_data=data, expected_network_data=data)


class TestCloudInitConfig(TestGenericConfig):
    CLASS = instance_config.CloudInitConfig

    def test_add_user(self):
        config = self.CLASS()
        config.add_user('admin')
        self._check(config, {},
                    {'users': [{'name': 'admin',
                                'groups': ['wheel']}]})

    def test_add_user_admin(self):
        config = self.CLASS()
        config.add_user('admin', admin=False)
        self._check(config, {},
                    {'users': [{'name': 'admin'}]})

    def test_add_user_sudo(self):
        config = self.CLASS()
        config.add_user('admin', sudo=True)
        self._check(config, {},
                    {'users': [{'name': 'admin',
                                'groups': ['wheel'],
                                'sudo': 'ALL=(ALL) NOPASSWD:ALL'}]})

    def test_add_user_passwd(self):
        config = self.CLASS()
        config.add_user('admin', password_hash='123')
        self._check(config, {},
                    {'users': [{'name': 'admin',
                                'groups': ['wheel'],
                                'passwd': '123'}]})

    def test_add_user_with_keys(self):
        config = self.CLASS(ssh_keys=['abc', 'def'])
        config.add_user('admin')
        self._check(config, {'public_keys': {'0': 'abc', '1': 'def'}},
                    {'users': [{'name': 'admin',
                                'groups': ['wheel'],
                                'ssh_authorized_keys': ['abc', 'def']}]})

    # Overriding tests since CloudInitConfig does not support plain strings
    # for user_data, only dictionaries.
    def test_custom_user_data(self):
        config = self.CLASS(user_data={'answer': 42})
        self._check(config, {}, {'answer': 42})

    def test_custom_user_data_with_users(self):
        config = self.CLASS(user_data={'answer': 42})
        config.add_user('admin')
        self._check(config, {},
                    {'users': [{'name': 'admin',
                                'groups': ['wheel']}],
                     'answer': 42})

    def test_user_data_not_dict(self):
        self.assertRaises(TypeError, self.CLASS, user_data="string")
        config = self.CLASS()
        config.user_data = "string"
        self.assertRaises(TypeError, config.populate_user_data)



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
from unittest import mock

from metalsmith_ansible.ansible_plugins.modules \
    import metalsmith_instances as mi

from metalsmith import exceptions as exc


class TestMetalsmithInstances(unittest.TestCase):

    @mock.patch('metalsmith.sources.detect', autospec=True)
    def test_get_source(self, mock_detect):
        mi._get_source({
            'image': {'href': 'overcloud-full'}
        })
        mi._get_source({
            'image': {
                'href': 'file://overcloud-full.qcow2',
                'checksum': 'asdf',
                'kernel': 'file://overcloud-full.vmlinuz',
                'ramdisk': 'file://overcloud-full.initrd'
            }
        })
        mock_detect.assert_has_calls([
            mock.call(
                image='overcloud-full',
                checksum=None,
                kernel=None,
                ramdisk=None
            ),
            mock.call(
                image='file://overcloud-full.qcow2',
                checksum='asdf',
                kernel='file://overcloud-full.vmlinuz',
                ramdisk='file://overcloud-full.initrd'
            )
        ])

    def test_reserve(self):
        provisioner = mock.Mock()
        instances = [{
            'name': 'node',
            'candidates': ['other_node'],
            'resource_class': 'boxen',
            'capabilities': {'foo': 'bar'},
            'traits': ['this', 'that'],
            'conductor_group': 'group'
        }, {}]
        reserved = [
            mock.Mock(id=1),
            mock.Mock(id=2),
        ]

        # test reserve success
        provisioner.reserve_node.side_effect = reserved

        result = mi.reserve(provisioner, instances, True)
        provisioner.reserve_node.assert_has_calls([
            mock.call(
                candidates=['other_node', 'node'],
                capabilities={'foo': 'bar'},
                conductor_group='group',
                resource_class='boxen',
                traits=['this', 'that']
            ),
            mock.call(
                candidates=None,
                capabilities=None,
                conductor_group=None,
                resource_class='baremetal',
                traits=None
            )
        ])
        self.assertTrue(result[0])
        self.assertEqual(reserved, result[1])

        # test reserve failure with cleanup
        instances = [{}, {}, {}]
        reserved = [
            mock.Mock(id=1),
            mock.Mock(id=2),
            exc.ReservationFailed('ouch')
        ]
        provisioner.reserve_node.side_effect = reserved
        self.assertRaises(exc.ReservationFailed, mi.reserve,
                          provisioner, instances, True)
        provisioner.unprovision_node.assert_has_calls([
            mock.call(1),
            mock.call(2)
        ])

    @mock.patch('metalsmith.sources.detect', autospec=True)
    @mock.patch('metalsmith.instance_config.CloudInitConfig', autospec=True)
    def test_provision(self, mock_config, mock_detect):
        config = mock_config.return_value
        image = mock_detect.return_value

        provisioner = mock.Mock()
        instances = [{
            'name': 'node-1',
            'hostname': 'overcloud-controller-1',
            'image': {'href': 'overcloud-full'}
        }, {
            'name': 'node-2',
            'hostname': 'overcloud-controller-2',
            'image': {'href': 'overcloud-full'},
            'nics': {'network': 'ctlplane'},
            'root_size_gb': 200,
            'swap_size_mb': 16,
            'netboot': True,
            'ssh_public_keys': 'abcd',
            'user_name': 'centos',
            'passwordless_sudo': False,
            'config_drive': {
                'meta_data': {'foo': 'bar'},
                'cloud_config': {'bootcmd': ['echo henlo world']}
            }
        }, {
            'name': 'node-3',
            'hostname': 'overcloud-controller-3',
            'image': {'href': 'overcloud-full'}
        }, {
            'name': 'node-4',
            'hostname': 'overcloud-compute-0',
            'image': {'href': 'overcloud-full'}
        }]
        provisioned = [
            mock.Mock(uuid=1),
            mock.Mock(uuid=2),
            mock.Mock(uuid=3),
            mock.Mock(uuid=4),
        ]

        # test provision success
        provisioner.provision_node.side_effect = provisioned

        # provision 4 nodes with concurrency of 2
        result = mi.provision(provisioner, instances, 3600, 2, True, True)
        provisioner.provision_node.assert_has_calls([
            mock.call(
                'node-1',
                config=config,
                hostname='overcloud-controller-1',
                image=image,
                netboot=False,
                nics=None,
                root_size_gb=None,
                swap_size_mb=None
            ),
            mock.call(
                'node-2',
                config=config,
                hostname='overcloud-controller-2',
                image=image,
                netboot=True,
                nics={'network': 'ctlplane'},
                root_size_gb=200,
                swap_size_mb=16
            ),
            mock.call(
                'node-3',
                config=config,
                hostname='overcloud-controller-3',
                image=image,
                netboot=False,
                nics=None,
                root_size_gb=None,
                swap_size_mb=None
            ),
            mock.call(
                'node-4',
                config=config,
                hostname='overcloud-compute-0',
                image=image,
                netboot=False,
                nics=None,
                root_size_gb=None,
                swap_size_mb=None
            ),
        ])
        mock_config.assert_has_calls([
            mock.call(ssh_keys=None, user_data=None, meta_data=None),
            mock.call(ssh_keys='abcd',
                      user_data={'bootcmd': ['echo henlo world']},
                      meta_data={'foo': 'bar'})
        ])
        config.add_user.assert_called_once_with(
            'centos', admin=True, sudo=False)
        mock_detect.assert_has_calls([
            mock.call(
                image='overcloud-full',
                checksum=None,
                kernel=None,
                ramdisk=None
            ),
            mock.call(
                image='overcloud-full',
                checksum=None,
                kernel=None,
                ramdisk=None
            ),
            mock.call(
                image='overcloud-full',
                checksum=None,
                kernel=None,
                ramdisk=None
            ),
            mock.call(
                image='overcloud-full',
                checksum=None,
                kernel=None,
                ramdisk=None
            ),
        ])
        self.assertTrue(result[0])
        self.assertEqual(provisioned, result[1])

        # test provision failure with cleanup
        instances = [{
            'name': 'node-1',
            'hostname': 'overcloud-controller-1',
            'image': {'href': 'overcloud-full'}
        }, {
            'name': 'node-2',
            'hostname': 'overcloud-controller-2',
            'image': {'href': 'overcloud-full'},
        }, {
            'name': 'node-3',
            'hostname': 'overcloud-controller-3',
            'image': {'href': 'overcloud-full'},
        }]
        provisioned = [
            mock.Mock(uuid=1),
            mock.Mock(uuid=2),
            exc.Error('ouch')
        ]
        provisioner.provision_node.side_effect = provisioned
        self.assertRaises(exc.Error, mi.provision,
                          provisioner, instances, 3600, 20, True, True)
        provisioner.unprovision_node.assert_has_calls([
            mock.call(1),
            mock.call(2)
        ])



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

from metalsmith import _network_metadata


class TestMetadataAdd(unittest.TestCase):

    def test_metadata_add_links(self):
        port = mock.Mock()
        network = mock.Mock()
        port.id = 'port_id'
        port.mac_address = 'aa:bb:cc:dd:ee:ff'
        network.mtu = 1500
        links = []
        expected = [{'id': 'port_id',
                     'type': 'phy',
                     'mtu': 1500,
                     'ethernet_mac_address': 'aa:bb:cc:dd:ee:ff'}]
        _network_metadata.metadata_add_links(links, port, network)
        self.assertEqual(expected, links)

    def test_metadata_add_services(self):
        subnet_a = mock.Mock()
        subnet_b = mock.Mock()
        subnet_a.dns_nameservers = ['192.0.2.1', '192.0.2.2']
        subnet_b.dns_nameservers = ['192.0.2.11', '192.0.2.22']
        subnets = [subnet_a, subnet_b]
        services = []
        expected = [{'address': '192.0.2.1', 'type': 'dns'},
                    {'address': '192.0.2.2', 'type': 'dns'},
                    {'address': '192.0.2.11', 'type': 'dns'},
                    {'address': '192.0.2.22', 'type': 'dns'}]
        _network_metadata.metadata_add_services(services, subnets)
        self.assertEqual(expected, services)

    def test_metadata_add_network_ipv4_dhcp(self):
        idx = 1
        fixed_ip = {'ip_address': '192.0.2.100', 'subnet_id': 'subnet_id'}

        port = mock.Mock()
        port.id = 'port_id'

        subnet = mock.Mock()
        subnet.cidr = '192.0.2.0/26'
        subnet.ip_version = 4
        subnet.is_dhcp_enabled = True
        subnet.host_routes = [
            {'destination': '192.0.2.64/26', 'nexthop': '192.0.2.1'},
            {'destination': '192.0.2.128/26', 'nexthop': '192.0.2.1'}
        ]
        subnet.dns_nameservers = ['192.0.2.11', '192.0.2.22']

        network = mock.Mock()
        network.id = 'network_id'
        network.name = 'net_name'

        networks = []
        expected = [{'id': 'net_name1',
                     'ip_address': '192.0.2.100',
                     'link': 'port_id',
                     'netmask': '255.255.255.192',
                     'network_id': 'network_id',
                     'routes': [{'gateway': '192.0.2.1',
                                 'netmask': '255.255.255.192',
                                 'network': '192.0.2.64'},
                                {'gateway': '192.0.2.1',
                                 'netmask': '255.255.255.192',
                                 'network': '192.0.2.128'}],
                     'services': [{'address': '192.0.2.11', 'type': 'dns'},
                                  {'address': '192.0.2.22', 'type': 'dns'}],
                     'type': 'ipv4_dhcp'}]
        _network_metadata.metadata_add_network(networks, idx, fixed_ip, port,
                                               network, subnet)
        self.assertEqual(expected, networks)

    def test_metadata_add_network_ipv6_stateful(self):
        idx = 1
        fixed_ip = {'ip_address': '2001:db8:1::10', 'subnet_id': 'subnet_id'}
        port = mock.Mock()
        port.id = 'port_id'

        subnet = mock.Mock()
        subnet.cidr = '2001:db8:1::/64'
        subnet.ip_version = 6
        subnet.ipv6_address_mode = 'dhcpv6-stateful'
        subnet.host_routes = [
            {'destination': '2001:db8:2::/64', 'nexthop': '2001:db8:1::1'},
            {'destination': '2001:db8:3::/64', 'nexthop': '2001:db8:1::1'}
        ]
        subnet.dns_nameservers = ['2001:db8:1::ee', '2001:db8:2::ff']

        network = mock.Mock()
        network.id = 'network_id'
        network.name = 'net_name'

        networks = []
        expected = [
            {'id': 'net_name1',
             'ip_address': '2001:db8:1::10',
             'link': 'port_id',
             'netmask': 'ffff:ffff:ffff:ffff::',
             'network_id': 'network_id',
             'routes': [{'gateway': '2001:db8:1::1',
                         'netmask': 'ffff:ffff:ffff:ffff::',
                         'network': '2001:db8:2::'},
                        {'gateway': '2001:db8:1::1',
                         'netmask': 'ffff:ffff:ffff:ffff::',
                         'network': '2001:db8:3::'}],
             'services': [{'address': '2001:db8:1::ee', 'type': 'dns'},
                          {'address': '2001:db8:2::ff', 'type': 'dns'}],
             'type': 'ipv6_dhcpv6-stateful'}]
        _network_metadata.metadata_add_network(networks, idx, fixed_ip, port,
                                               network, subnet)
        self.assertEqual(expected, networks)



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import logging
import re
import sys

from openstack import exceptions as os_exc

from metalsmith import exceptions


LOG = logging.getLogger(__name__)


def log_res(res):
    if res is None:
        return None
    elif getattr(res, 'name', None):
        return '%s (UUID %s)' % (res.name, res.id)
    else:
        return res.id


def get_capabilities(node):
    caps = node.properties.get('capabilities') or {}
    if not isinstance(caps, dict):
        caps = dict(x.split(':', 1) for x in caps.split(',') if x)
    return caps


def get_root_disk(root_size_gb, node):
    """Validate and calculate the root disk size."""
    if root_size_gb is not None:
        if not isinstance(root_size_gb, int):
            raise TypeError("The root_size_gb argument must be "
                            "a positive integer, got %r" % root_size_gb)
        elif root_size_gb <= 0:
            raise ValueError("The root_size_gb argument must be "
                             "a positive integer, got %d" % root_size_gb)
    else:
        try:
            assert int(node.properties['local_gb']) > 0
        except KeyError:
            LOG.debug('No local_gb for node %s and no root partition size '
                      'specified', log_res(node))
            return
        except (TypeError, ValueError, AssertionError):
            raise exceptions.UnknownRootDiskSize(
                'The local_gb for node %(node)s is invalid: '
                'expected positive integer, got %(value)s' %
                {'node': log_res(node),
                 'value': node.properties['local_gb']})

        # allow for partitioning and config drive
        root_size_gb = int(node.properties['local_gb']) - 1

    return root_size_gb


_HOSTNAME_RE = re.compile(r"""^
[a-z0-9][a-z0-9\-]{0,61}[a-z0-9]      # host
(\.[a-z0-9][a-z0-9\-]{0,61}[a-z0-9])* # domain
$""", re.IGNORECASE | re.VERBOSE)


def is_hostname_safe(hostname):
    """Check for valid host name.

    Nominally, checks that the supplied hostname conforms to:
        * http://en.wikipedia.org/wiki/Hostname
        * http://tools.ietf.org/html/rfc952
        * http://tools.ietf.org/html/rfc1123

    :param hostname: The hostname to be validated.
    :returns: True if valid. False if not.
    """
    if not isinstance(hostname, str) or len(hostname) > 255:
        return False

    return _HOSTNAME_RE.match(hostname) is not None


def check_hostname(hostname):
    """Check the provided host name.

    :raises: ValueError on inappropriate value of ``hostname``
    """
    if hostname is not None and not is_hostname_safe(hostname):
        raise ValueError("%s cannot be used as a hostname" % hostname)


def parse_checksums(checksums):
    """Parse standard checksums file."""
    result = {}
    for line in checksums.split('\n'):
        if not line.strip():
            continue

        checksum, fname = line.strip().split(None, 1)
        result[fname.strip().lstrip('*')] = checksum.strip()

    return result


def default_hostname(node):
    if node.name and is_hostname_safe(node.name):
        return node.name
    else:
        return node.id


def hostname_for(node, allocation=None):
    if allocation is not None and allocation.name:
        return allocation.name
    else:
        return default_hostname(node)


@contextlib.contextmanager
def reraise_os_exc(reraise_as, failure_message='Clean up failed'):
    exc_info = sys.exc_info()
    is_expected = isinstance(exc_info[1], os_exc.SDKException)

    try:
        yield is_expected
    except Exception:
        LOG.exception(failure_message)

    if is_expected:
        raise reraise_as(str(exc_info[1]))
    else:
        raise exc_info[1]



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from concurrent import futures
import io
import logging

from ansible.module_utils.basic import AnsibleModule
try:
    from ansible.module_utils.openstack import openstack_cloud_from_module
    from ansible.module_utils.openstack import openstack_full_argument_spec
    from ansible.module_utils.openstack import openstack_module_kwargs
except ImportError:
    openstack_cloud_from_module = None
    openstack_full_argument_spec = None
    openstack_module_kwargs = None

import metalsmith
from metalsmith import instance_config
from metalsmith import sources

import yaml


DOCUMENTATION = '''
---
module: metalsmith_instances
short_description: Manage baremetal instances with metalsmith
author: "Steve Baker (@stevebaker)"
description:
  - Provision and unprovision ironic baremetal instances using metalsmith,
    which is a a simple tool to provision bare metal machines using
    OpenStack Bare Metal Service (ironic) and, optionally, OpenStack
    Image Service (glance) and OpenStack Networking Service (neutron).
options:
  instances:
    description:
      - List of node description dicts to perform operations on
    type: list
    default: []
    elements: dict
    suboptions:
      hostname:
        description:
          - Host name to use, defaults to Node's name or UUID
        type: str
      name:
        description:
          - The name of an existing node to provision, this name is appended
            to the candidates list
        type: str
      candidates:
        description:
          - List of nodes (UUIDs or names) to be considered for deployment
        type: list
        elements: str
      image:
        description:
          - Details of the image you want to provision onto the node
        type: dict
        required: True
        suboptions:
          href:
            description:
              - Image to use (name, UUID or URL)
            type: str
            required: True
          checksum :
            description:
              - Image MD5 checksum or URL with checksums
            type: str
          kernel:
            description:
              - URL of the image's kernel
            type: str
          ramdisk:
            description:
              - URL of the image's ramdisk
            type: str
      nics:
        description:
          - List of requested NICs
        type: list
        elements: dict
        suboptions:
          network:
            description:
              - Network to create a port on (name or UUID)
          subnet:
            description:
              - Subnet to create a port on (name or UUID)
          port:
            description:
              - Port to attach (name or UUID)
          fixed_ip:
            description:
              - Attach IP from the network

      netboot:
        description:
          - Boot from network instead of local disk
        default: no
        type: bool
      root_size_gb:
        description:
          - Root partition size (in GiB), defaults to (local_gb - 1)
        type: int
      swap_size_mb:
        description:
          - Swap partition size (in MiB), defaults to no swap
        type: int
      capabilities:
        description:
          - Selection criteria to match the node capabilities
        type: dict
      traits:
        description:
          - Traits the node should have
        type: list
        elements: str
      ssh_public_keys:
        description:
          - SSH public keys to load
      resource_class:
        description:
          - Node resource class to provision
        type: str
        default: baremetal
      conductor_group:
        description:
          - Conductor group to pick the node from
        type: str
      user_name:
        description:
          - Name of the admin user to create
        type: str
      passwordless_sudo:
        description:
          - Allow password-less sudo for the user
        default: yes
        type: bool
      config_drive:
        description:
          - Extra data to add to the config-drive generated for this instance
        type: dict
        suboptions:
          cloud_config:
            description:
              - Dict of cloud-init cloud-config tasks to run on node
                boot. The 'users' directive can be used to configure extra
                users other than the 'user_name' admin user.
            type: dict
          meta_data:
            description:
              - Extra metadata to include with the config-drive metadata.
                This will be added to the generated metadata
                'public_keys', 'uuid', 'name', and 'hostname'.
            type: dict
  clean_up:
    description:
      - Clean up resources on failure
    default: yes
    type: bool
  state:
    description:
      - Desired provision state, "present" to provision,
        "absent" to unprovision, "reserved" to create an allocation
        record without changing the node state
    default: present
    choices:
    - present
    - absent
    - reserved
  wait:
    description:
      - A boolean value instructing the module to wait for node provision
        to complete before returning.  A 'yes' is implied if the number of
        instances is more than the concurrency.
    type: bool
    default: no
  timeout:
    description:
      - An integer value representing the number of seconds to wait for the
        node provision to complete.
    type: int
    default: 3600
  concurrency:
    description:
      - Maximum number of instances to provision at once. Set to 0 to have no
        concurrency limit
    type: int
    default: 0
  log_level:
    description:
      - Set the logging level for the log which is available in the
        returned 'logging' result.
    default: info
    choices:
    - debug
    - info
    - warning
    - error
'''

EXAMPLES = '''
- name: Provision instances
  metalsmith_instances:
    instances:
    - name: node-0
      hostname: compute-0
      image: overcloud-full
    state: present
    wait: true
    clean_up: false
    timeout: 1200
    concurrency: 20
    log_level: info
  register: baremetal_provisioned

- name: Metalsmith log for provision instances
  debug:
    var: baremetal_provisioned.logging
'''


METALSMITH_LOG_MAP = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR
}

BASE_LOG_MAP = {
    'debug': logging.INFO,
    'info': logging.WARNING,
    'warning': logging.WARNING,
    'error': logging.ERROR
}


def _get_source(instance):
    image = instance.get('image')
    return sources.detect(image=image.get('href'),
                          kernel=image.get('kernel'),
                          ramdisk=image.get('ramdisk'),
                          checksum=image.get('checksum'))


def reserve(provisioner, instances, clean_up):
    nodes = []
    for instance in instances:
        candidates = instance.get('candidates', [])
        if instance.get('name') is not None:
            candidates.append(instance['name'])
        if not candidates:
            candidates = None
        try:
            node = provisioner.reserve_node(
                resource_class=instance.get('resource_class', 'baremetal'),
                capabilities=instance.get('capabilities'),
                candidates=candidates,
                traits=instance.get('traits'),
                conductor_group=instance.get('conductor_group')),
            if isinstance(node, tuple):
                node = node[0]
            nodes.append(node)
            # side-effect of populating the instance name, which is passed to
            # a later provision step
            instance['name'] = node.id
        except Exception as exc:
            if clean_up:
                # Remove all reservations on failure
                _release_nodes(provisioner, [i.id for i in nodes])
            raise exc
    return len(nodes) > 0, nodes


def _release_nodes(provisioner, node_ids):
    for node in node_ids:
        try:
            provisioner.unprovision_node(node)
        except Exception:
            pass


def provision(provisioner, instances, timeout, concurrency, clean_up, wait):
    if not instances:
        return False, []

    # first, ensure all instances are reserved
    reserve(provisioner, [i for i in instances if not i.get('name')], clean_up)

    nodes = []

    # no limit on concurrency, create a worker for every instance
    if concurrency < 1:
        concurrency = len(instances)

    # if concurrency is less than instances, need to wait for
    # instance completion
    if concurrency < len(instances):
        wait = True

    provision_jobs = []
    exceptions = []
    with futures.ThreadPoolExecutor(max_workers=concurrency) as p:
        for i in instances:
            provision_jobs.append(p.submit(
                _provision_instance, provisioner, i, nodes, timeout, wait
            ))
    for job in futures.as_completed(provision_jobs):
        e = job.exception()
        if e:
            exceptions.append(e)

            if clean_up:
                # first, cancel all jobs
                for job in provision_jobs:
                    job.cancel()
                # Unprovision all provisioned so far.
                # This is best-effort as some provision calls may have
                # started but not yet appended to nodes.
                _release_nodes(provisioner, [i.uuid for i in nodes])
                nodes = []
    if exceptions:
        # TODO(sbaker) future enhancement to tolerate a proportion of failures
        # so that provisioning and deployment can continue
        raise exceptions[0]

    return len(nodes) > 0, nodes


def _provision_instance(provisioner, instance, nodes, timeout, wait):
    name = instance.get('name')

    image = _get_source(instance)
    ssh_keys = instance.get('ssh_public_keys')
    config_drive = instance.get('config_drive', {})
    cloud_config = config_drive.get('cloud_config')
    meta_data = config_drive.get('meta_data')
    config = instance_config.CloudInitConfig(ssh_keys=ssh_keys,
                                             user_data=cloud_config,
                                             meta_data=meta_data)
    if instance.get('user_name'):
        config.add_user(instance.get('user_name'), admin=True,
                        sudo=instance.get('passwordless_sudo', True))
    node = provisioner.provision_node(
        name,
        config=config,
        hostname=instance.get('hostname'),
        image=image,
        nics=instance.get('nics'),
        root_size_gb=instance.get('root_size_gb'),
        swap_size_mb=instance.get('swap_size_mb'),
        netboot=instance.get('netboot', False)
    )
    nodes.append(node)
    if wait:
        provisioner.wait_for_provisioning(
            [node.uuid], timeout=timeout)


def unprovision(provisioner, instances):
    for instance in instances:
        provisioner.unprovision_node(instance.get('name'))
    return True


def _configure_logging(log_level):
    log_fmt = ('%(asctime)s %(levelname)s %(name)s: %(message)s')
    urllib_level = logging.CRITICAL

    base_level = BASE_LOG_MAP[log_level]
    metalsmith_level = METALSMITH_LOG_MAP[log_level]

    logging.basicConfig(level=base_level, format=log_fmt)
    logging.getLogger('urllib3.connectionpool').setLevel(urllib_level)
    logger = logging.getLogger('metalsmith')
    logger.setLevel(metalsmith_level)
    log_stream = io.StringIO()
    logger.addHandler(logging.StreamHandler(log_stream))
    return log_stream


def main():
    if not openstack_full_argument_spec:
        raise RuntimeError(
            'This module requires ansible-collections-openstack')

    argument_spec = openstack_full_argument_spec(
        **yaml.safe_load(DOCUMENTATION)['options']
    )
    module_kwargs = openstack_module_kwargs()
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=False,
        **module_kwargs
    )

    log_stream = _configure_logging(module.params['log_level'])

    try:
        sdk, cloud = openstack_cloud_from_module(module)
        provisioner = metalsmith.Provisioner(cloud_region=cloud.config)
        instances = module.params['instances']
        state = module.params['state']
        concurrency = module.params['concurrency']
        timeout = module.params['timeout']
        wait = module.params['wait']
        clean_up = module.params['clean_up']

        if state == 'present':
            changed, nodes = provision(provisioner, instances,
                                       timeout, concurrency, clean_up,
                                       wait)
            instances = [{
                'name': i.node.name or i.uuid,
                'hostname': i.hostname,
                'id': i.uuid,
            } for i in nodes]
            module.exit_json(
                changed=changed,
                msg="{} instances provisioned".format(len(nodes)),
                instances=instances,
                logging=log_stream.getvalue()
            )

        if state == 'reserved':
            changed, nodes = reserve(provisioner, instances, clean_up)
            module.exit_json(
                changed=changed,
                msg="{} instances reserved".format(len(nodes)),
                ids=[node.id for node in nodes],
                instances=instances,
                logging=log_stream.getvalue()
            )

        if state == 'absent':
            changed = unprovision(provisioner, instances)
            module.exit_json(
                changed=changed,
                msg="{} nodes unprovisioned".format(len(instances)),
                logging=log_stream.getvalue()
            )
    except Exception as e:
        module.fail_json(
            msg=str(e),
            logging=log_stream.getvalue()
        )


if __name__ == '__main__':
    main()
